import {ArrayUtils} from 'great-jsutils';

export class ZorroTableTreeUtils<T> {

    keys: any;
    data: any;
    expandDataCache: any;

    constructor({
                    keys = {
                        idKey: "id",
                        pIdKey: "parentId",
                        pKey: "parent",
                        childKey: "children"
                    }, data = {}, expandDataCache={}
                }={}) {
        this.keys = keys;
        this.data = data;
        this.expandDataCache = expandDataCache;
    }

    init(): void {
        this.data.forEach(item => {
            this.expandDataCache[item.key] = this.convertTreeToList(item);
        });
    }

    /*展开与关闭*/
    collapse(array: Array<T>, data: T, $event: boolean): void {
        if ($event === false) {
            if (data && data[this.keys.childKey]) {
                data[this.keys.childKey].forEach(d => {
                    // @ts-ignore
                    const target = array.find(a => a[this.keys.idKey] === d[this.keys.idKey]);
                    if (target) {
                        target["expand"] = false;
                    }
                    this.collapse(array, target, false);
                });
            }
        }
    }

    /*添加节点*/
    toAddNode(newNodes: Array<T>) {
        if (ArrayUtils.validArray(newNodes)) {
            let parentId = newNodes[0][this.keys.pIdKey];
            let parentNode = this.getCacheNode(parentId);
            this.addNodes(parentNode, newNodes);
        }
    }

    /*更新节点*/
    toUpdateNode(newNode: T) {
        if(!newNode){
            return;
        }
        let id = newNode[this.keys.idKey];
        let oldNode = this.getCacheNode(id);
        ArrayUtils.update([oldNode], newNode, this.keys);
        ArrayUtils.update(this.data, newNode, this.keys);
    }

    /*移除节点*/
    toRemoveNode(node) {
        if(!node){
            return;
        }
        if(this.expandDataCache[node[this.keys.idKey]]){
            delete this.expandDataCache[node[this.keys.idKey]];
            return;
        }
        if(node[this.keys.childKey] && ArrayUtils.validArray(node[this.keys.childKey])){
            node[this.keys.childKey].forEach(item=>{
                this.toRemoveNode(item);
            });
        }
        let id = node[this.keys.idKey];
        let oldNode = this.getCacheNode(id);
        if(oldNode){
            let rootKey = this.getRootKey(oldNode);
            ArrayUtils.remove(this.expandDataCache[rootKey], node, this.keys);
            ArrayUtils.remove(this.data, node, this.keys);
        }
    }

    /* 获取父节点 */
    private getCacheNode(id: any) {
        for (let key in this.expandDataCache) {
            let cacheNode = this.expandDataCache[key].find((v, i, a) => {
                if (v.key == id) {
                    return v
                }
            });
            if (cacheNode) {
                return cacheNode;
            }
        }
        return null;
    }

    /*批量向父节点中追加节点*/
    addNodes(parentNode: T, newNodes: Array<any>) {
        for (let i = newNodes.length - 1; i > -1; i--) {
            let newNode = newNodes[i];
            if(!ArrayUtils.validArray(parentNode[this.keys.childKey] )){
                parentNode[this.keys.childKey] = [] as T[];
            }
            newNode[this.keys.pIdKey] = parentNode[this.keys.idKey];
            if (!this.expandDataCache[this.keys.idKey]) {
                parentNode[this.keys.childKey].push(newNode as T);
                ArrayUtils.add(this.data, newNode, this.keys);
                newNode[this.keys.pKey] = parentNode;
                let index = this.expandDataCache[this.getRootKey(parentNode)].findIndex((v, i, a) => {
                    if (v[this.keys.idKey] == parentNode[this.keys.idKey]) {
                        return i
                    };
                });
                newNode.level = parentNode["level"] + 1;
                this.expandDataCache[this.getRootKey(parentNode)].splice(index + 1, 0, newNode);
            }
        }
    }

    /* 获取一级节点的key*/
    private getRootKey(data: T) {
        if (data && data[this.keys.pKey]) {
            return this.getRootKey(data[this.keys.pKey]);
        } else {
            return data[this.keys.idKey];
        }
    }

    private convertTreeToList(root: object): T[] {
        const stack = [];
        const array = [];
        const hashMap = {};
        stack.push({...root, level: 0, expand: false});

        while (stack.length !== 0) {
            const node = stack.pop();
            this.visitNode(node, hashMap, array);
            if (node.children) {
                for (let i = node.children.length - 1; i >= 0; i--) {
                    stack.push({...node.children[i], level: node.level + 1, expand: false, parent: node});
                }
            }
        }

        return array;
    }

    private visitNode(node: T, hashMap: object, array: T[]): void {
        if (!hashMap[node[this.keys.idKey]]) {
            hashMap[node[this.keys.idKey]] = true;
            array.push(node);
        }
    }
}

